// Code generated by "stringer -type deviceState -trimprefix=deviceState"; DO NOT EDIT.

package device

import "strconv"

const _deviceState_name = "DownUpClosed"

var _deviceState_index = [...]uint8{0, 4, 6, 12}

func (i deviceState) String() string {
	if i >= deviceState(len(_deviceState_index)-1) {
		return "deviceState(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _deviceState_name[_deviceState_index[i]:_deviceState_index[i+1]]
}
